--[[doc
h1.	API: cargBags Core

*Description:*
	The main core used for bag organizing and filtering, it is the connection between handler, plugins and layout

*Spawn with:*
*	cargBags:Spawn(name, parentFrame): Create a new bag object

*cargBags functions:*
*	:Spawn(name, parentFrame): Create a new bag object
*	:RegisterStyle(name, func): Register a new layout for use
*	:SetActiveLayout(name): Set the active layout
*	:RegisterHandler(name, handler, ...): Register a new handler, additional options are passed to handler:Enable()
*	:SetActiveHandler(name, ...): Set the active handler, additional options are passed to handler:Enable()
*	:GetHandler(): Returns the currently active handler
*	:RegisterPlugin(name, func): Registers a new plugin, the function is called when the plugin needs to be created
*	:AddCallback(object, func): Registers a callback for an object to be notified when the bags are updated
*	table, min, max = :ParseBags(bags): Returns the parsed bags from this table or string
.p)			bags = "bags": main bags (1-4)
			bags = "backpack+bags": default inventory (0-4)
			bags = "bank": bank bags (5-11)
			bags = "bagpack": backpack (0)
			bags = "bankframe": bankframe (-1)
			bags = "keyring": keyring (-2)
			bags = "X-Y": bags with id's from 'X' to 'Y'
			bags = X: bagID of single bag
			bags = {} : indexed array with all bagID's used

*cargBags callback functions:*
*	:PreUpdateBags(event, bagID, slotID): fired before the objects are updated (filters added, item buttons updated)
*	:PostUpdateBags(event, bagID, slotID): fired after the objects are updated (filters added, item buttons updated)
*	:PreCheckFilters(item, updateType): fired before the filters of an item are checked

*Bag object functions:*
*	:SpawnPlugin(name, ...): Spawn a plugin for use in the layout, parameters are plugin-specific
*	:IterateButtons(): Creates an iterator over all item buttons for use within a for-loop [overwritable]
*	:RequestItemData(): Fetch the item's main information into the item-table (for use within UpdateButton-functions)
*	:RequestItemData(item, bagID, slotID): Fetches the information into a custom table
*	:RequestCooldownData(): Fetch the item's cooldown information into the item-table (for use within UpdateButton-functions)
*	:RequestCooldownData(): Fetch the item's cooldown information into a custom table
*	:GetHandler(): Returns the currently active handler, provided from cargBags for convenience
*	:CheckFilters(item): Returns if the item table fits into this bag object [overwritable]
*	:SetFilter(func, state): Sets the state for the filter function (true: enabled, nil: disabled, -1: inverted; Default: nil)

*Bag object callback functions:*
*	:PostCreateBag(bag, bagID): fired after a bag was created in  the object
*	:PostAddButton(button): fired after a button was added/moved to the object
*	:PostRemoveButton(button): fired after a button was removed from the object
*	:UpdateButton(button, item, updateType): fired if a button needs to be updated
*	:UpdateButtonLock(button, i, updateType): fired if a button's lock needs to be updated
*	:UpdateButtonCooldown(button, i, updateType): fired if a button's' cooldown needs to be updated
*	:UpdateButtonPositions(): fired if the button positions need to be updated, typically after adding/removing buttons from the object
doc]]

local _G = getfenv(0)
local select = select
local type = type

-- add-on object
local cargBags = CreateFrame("Frame", "cargBags")

local Prototype = CreateFrame"Button"
local metatable = {__index = Prototype}

local print = function(a) ChatFrame1:AddMessage("|cffee8800cargBags:|r "..tostring(a)) end
local error = function(...) print("|cffff0000Error:|r "..string.format(...)) end

local styles, style = {}
local objects = {}
local bags = {}
local handlers, handler = {}
local tempButtons = {}
local callbacks = {}
local plugins = {}
local init = true



--[[##############################
	Callback functions
		Because programmers are lazy
################################]]

-- Tries to call an object's function if it exists
local function fire(object, callback, ...)
	if(not object or not object[callback]) then return end
	--debug(object.GetName and object:GetName(), callback)
	object[callback](object, ...)
end

-- Tries to call the functions of all objects
local function fireAll(callback, ...)
	for _, object in ipairs(objects) do
		fire(object, callback, ...)
	end
end

-- Calls the functions from the callbacks-table
local function fireCallbacks(...)
	for object, func in pairs(callbacks) do
		func(object, ...)
	end
end



--[[##############################
	Registering/enabling functions
		Here you can register styles, plugins and handlers
################################]]

-- Register a style
function cargBags:RegisterStyle(name, func)
	if(type(name) ~= "string") then return error("Bad argument #1 to 'RegisterStyle' (string expected, got %s)", type(name)) end
	if(type(func) ~= "table" and type(getmetatable(func).__call) ~= "function") then return error("Bad argument #2 to 'RegisterStyle' (table expected, got %s)", type(func)) end
	if(styles[name]) then return error("Style [%s] already registered.", name) end
	if(not style) then style = name end

	styles[name] = func
end

-- Set the active style
function cargBags:SetActiveStyle(name, noVerbose)
	if(type(name) ~= "string") then return error("Bad argument #1 to 'SetActiveStyle' (string expected, got %s)", type(name)) end
	if(not styles[name]) then return not noVerbose and error("Style [%s] does not exist.", name) end

	style = name
	return true
end

-- Register a handler
function cargBags:RegisterHandler(name, func, ...)
	if(type(name) ~= "string") then return error("Bad argument #1 to 'RegisterHandler' (string expected, got %s)", type(name)) end
	if(type(func) ~= "table") then return error("Bad argument #2 to 'RegisterHandler' (table expected, got %s)", type(func)) end
	if(handlers[name]) then return error("Handler [%s] already registered.", name) end
	if(not handler) then
		handler = func
		fire(handler, "Enable", ...)
	end

	handlers[name] = func
end

-- Set the active handler
function cargBags:SetActiveHandler(name, noVerbose, ...)
	if(type(name) ~= "string") then return error("Bad argument #1 to 'SetActiveHandler' (string expected, got %s)", type(name)) end
	if(not handlers[name]) then return not noVerbose and error("Handler [%s] does not exist.", name) end

	fire(handler, "Disable")
	handler = handlers[name]
	fire(handler, "Enable", ...)
	if(not init) then self:UpdateBags() end
	return true
end

-- Register a plugin
function cargBags:RegisterPlugin(name, func)
	if(type(name) ~= "string") then return error("Bad argument #1 to 'RegisterPlugin' (string expected, got %s)", type(name)) end
	if(type(func) ~= "function") then return error("Bad argument #2 to 'RegisterPlugin' (function expected, got %s)", type(func)) end
	if(plugins[name]) then return error("Plugin [%s] already registered.", name) end
	plugins[name] = func
end

-- Spawn a plugin
function cargBags:SpawnPlugin(name, ...)
	if(type(name) ~= "string") then return error("Bad argument #1 to 'SpawnPlugin' (string expected, got %s)", type(name)) end
	local plugin
	if(plugins[name]) then
		plugin = plugins[name](self, ...)
	end
	if(plugin and not self[name]) then self[name] = plugin end
	return plugin
end

-- Add a callback for UpdateBags()
function cargBags:AddCallback(object, func)
	if(type(object) ~= "table") then return error("Bad argument #1 to 'AddCallback' (table expected, got %s)", type(name)) end
	if(type(func) ~= "function") then return error("Bad argument #2 to 'AddCallback' (function expected, got %s)", type(func)) end
	callbacks[object] = func
end

-- Get the currently active handler
function cargBags:GetHandler()
	return handler
end

-- Spawn a new bag object
function cargBags:Spawn(name, ...)
		
	if(not name) then return error("Unable to create frame. No name was defined.") end
	if(not style) then return error("Unable to create frame. No styles have been registered.") end
	if(not handler) then return error("Unable to create frame. No handlers have been registered.") end

	local style = styles[style]
	local object = CreateFrame("Button", name, parent or UIParent)

	object = setmetatable(object, metatable)

	object.Name = name
	object.Bags = {}
	object.Buttons = {}
	object.NumButtons = 0

	object.Filters = {}
	object.updateNeeded = true
	object.Init = true

	style(object, name, ...)
	fire(handler, "Init", object, name, ...)

	objects[#objects+1] = object

	return object
end

local bankTable
-- Some helper function for standardized bag-strings
function cargBags:ParseBags(bags)
	if(type(bags) == "table") then return bags end
	if(bags == "backpack+bags") then return nil, 0, 4 end
	if(bags == "bankframe+bank") then bankTable = bankTable or {-1, 5, 6, 7, 8, 9, 10, 11} return bankTable end
	if(bags == "bags") then return nil, 1, 4 end
	if(bags == "bank") then return nil, 5, 11 end
	if(bags == "backpack") then return nil, 0 end
	if(bags == "bankframe") then return nil, -1 end
	if(bags == "keyring") then return nil, -2 end
	local min, max = bags and bags:match("(%d+)-(%d+)")
	if(min) then return nil, min, max end
	return nil, tonumber(bags)
end



--[[##############################
	Button-recycling functions
		where the buttons are created, moved and deleted
################################]]

-- Remove a button and insert it in the temporary storage
local function recycleButton(button)
	local tpl, bagID, slotID = button.template, button.bagID, button.slotID
	if(not bags[bagID]) then bags[bagID] = {} end
	bags[bagID][slotID] = nil
	button:Hide()
	if(not tempButtons[tpl]) then tempButtons[tpl] = {} end
	tinsert(tempButtons[tpl], button)
end

-- Create a button with the defined template or
-- fetch it from the temporary storage
local slots = 0
local function createButton(tpl, bagID, slotID)
	local button
	if(tempButtons[tpl] and #tempButtons[tpl] > 0) then
		button = tremove(tempButtons[tpl])
	else
		slots = slots+1
		local name = "cargBagsSlot"..slots
		button = handler:CreateButton(tpl, name)
		button.Name = name
		button.template = tpl
	end

	button:SetID(slotID)
	button.slotID = slotID
	button.bagID = bagID
	bags[bagID][slotID] = button

	return button
end

-- Create a 'parent bag' for the bag object
local function createObjectBag(object, bagID)
	local bag = CreateFrame("Frame", nil, object)
	bag.Size = 0
	bag:SetID(bagID)
	object.BagsLowest = object.BagsLowest == nil and bagID or min(object.BagsLowest or 0, bagID)
	object.BagsHighest = object.BagsHighest == nil and bagID or max(object.BagsHighest or 0, bagID)
	object.Bags[bagID] = bag
	fire(object, "PostCreateBag", bag, bagID)
	return bag
end

-- Move a button to the specified bag object or
-- recycle it if the object is nil
local function move(button, object)
	local bagID, slotID = button.bagID, button.slotID
	local oldObject = button.Object

	if(oldObject) then
		if(oldObject == object) then return true end
		oldObject.Buttons[button] = nil
		oldObject.NumButtons = oldObject.NumButtons - 1
		oldObject.updateNeeded = true
		fire(oldObject, "PostRemoveButton", button)
	end

	button.Object = object
	if(not object) then return recycleButton(button) end

	local bag = object.Bags[bagID] or createObjectBag(object, bagID)
	button:SetParent(bag)
	button:Show()
	object.Buttons[button] = true
	object.NumButtons = object.NumButtons + 1
	object.updateNeeded = true
	fire(object, "PostAddButton", button)
	return true
end

-- get the right button from the button table or create one
local function getButton(bagID, slotID, noCreate)
	local tpl = handler:GetButtonTemplateName(bagID, slotID)
	if(not bags[bagID]) then bags[bagID] = {} end
	if(bags[bagID][slotID]) then
		local button = bags[bagID][slotID]
		if(button.template == tpl) then
			return button
		else
			move(button, nil)
		end
	end
	return not noCreate and createButton(tpl, bagID, slotID)
end



--[[##############################
	Core filter logic
		Check for filters and update slots
################################]]

-- Ask all objects if the button can go into them
-- and move it to the first one where it fits
local function checkButtonObject(i)
	local bagID = i.bagID
	for _, object in ipairs(objects) do
		if(not (object.Bags[bagID] and object.Bags[bagID].Hidden) and object:CheckFilters(i)) then
			return move(getButton(bagID, i.slotID), object)
		end
	end
	return nil
end

local i = {}
local LOCK, CD = 1, 2
-- First fetch the item data from the handler, then check if the filters
-- are right, and then let the layout update the button
local function updateSlot(bagID, slotID, updateType)
	i.bagID, i.slotID = bagID, slotID
	if(not updateType or updateType == LOCK) then 
		handler:LoadItemData(i, updateType)
	end
	if(not updateType or updateType == CD) then
		handler:LoadCooldownData(i)
	end

	fire(cargBags, "PreCheckFilters", i, updateType)
	local button = getButton(bagID, slotID, true)
	local object = button and button.Object
	if(object and object.Bags[bagID].Hidden) then
		return move(button, nil)
	end

	if(updateType and object or checkButtonObject(i)) then
		button = getButton(bagID, slotID, true)
		object = button.Object
		if(not updateType) then fire(object, "UpdateButton", button, i, updateType) end
		if(updateType ~= CD) then fire(object, "UpdateButtonLock", button, i, updateType) end
		if(updateType ~= LOCK) then fire(object, "UpdateButtonCooldown", button, i, updateType) end
		fire(object, "PostUpdateButton", button, i, updateType)
	elseif(button) then
		move(button, nil)
	end
end

-- Update all slots of this bag
local function updateBag(bagID, updateType)
	if(not bags[bagID]) then bags[bagID] = {} end
	local bag = bags[bagID]
	local prevSlots =  bag.slots or 0
	local numSlots = handler.GetContainerNumSlots(bagID)
	_, bag.bagType = handler.GetContainerNumFreeSlots(bagID)
	bag.slots = numSlots
	for slotID=1, max(prevSlots, numSlots) do
		if(slotID <= numSlots) then
			updateSlot(bagID, slotID, updateType)
		elseif(bag[slotID]) then
			move(bag[slotID], nil)
		end
	end
end



--[[##############################
	The main update function
		Here begins and ends the function-switching
################################]]

local bagclosed
local timedBags = {}
-- Update bags one frame later
cargBags:SetScript("OnUpdate", function(self)
	self:Hide()
	self:UpdateBags("CB_TIMED")
end)

-- This activates the update checking
local function timeBag(bagID)
	timedUpdate = true
	timedBags[bagID] = true
	cargBags:Show()
end

-- The main update routine, handles ... umm... everything
function cargBags:UpdateBags(event, bagID, slotID)
	if(init) then init = nil end

	-- Some event handling
	local updateType
	-- Update either the bank slot or the bank bag
	if(event == "PLAYERBANKSLOTS_CHANGED") then
		if(bagID <= NUM_BANKGENERIC_SLOTS) then
			slotID = bagID
			bagID = -1
		else
			return timeBag(bagID-NUM_BANKGENERIC_SLOTS)
		end
	-- Only update item lock info
	elseif(event == "ITEM_LOCK_CHANGED") then
		updateType = LOCK
	-- Only update cooldown info
	elseif(event == "BAG_UPDATE_COOLDOWN") then
		updateType = CD
	-- Only update callbacks
	elseif(event == "PLAYER_MONEY") then
		return fireCallbacks(event, bagID, slotID)
	-- schedule updating after closed bag
	elseif(event == "BAG_CLOSED") then
		bagclosed = bagID
		return
	elseif(bagclosed) then
		bagclosed, bagID, slotID = nil
	-- fake event for hiding bags
	elseif(event == "CB_BAG_HIDDEN") then
		bagID.updateNeeded = true -- Well, it's actually an object
		bagID = slotID
		slotID = nil
	elseif(event == "BAG_UPDATE" and bagID) then
		return timeBag(bagID)
	elseif(event == "PLAYER_LOGIN") then
		timeBag(0)
		timeBag(-2)
		return fireCallbacks(event, bagID, slotID)
	end

	-- Now start with updating the bags
	fireAll("PreUpdateBags", event, bagID, slotID)

	if(bagID and slotID) then
		updateSlot(bagID, slotID, updateType)
	elseif(bagID) then
		updateBag(bagID, updateType)
	elseif(timedUpdate) then
		for id, update in pairs(timedBags) do
			if(update) then
				updateBag(id)
				timedBags[id] = nil
			end
		end
		timedUpdate = nil
	else
		for i = -2, (NUM_BAG_SLOTS+NUM_BANKBAGSLOTS) do
			updateBag(i, updateType)
		end
	end

	-- Lots of PostUpdate-callbacks
	for _, object in ipairs(objects) do
		if(object.updateNeeded or object.PositionEveryTime) then fire(object, "UpdateButtonPositions", event, bagID, slotID) object.updateNeeded = nil end
		fire(object, "PostUpdateBags", event, bagID, slotID)
		if(object.Init) then object.Init = nil end
	end

	-- The plugins want, too
	fireCallbacks(event, bagID, slotID)
end



--[[##############################
	Functions of the bag object
		THEY R EVIL!!!!1111
################################]]

-- Check if the passed item data can go into this object
function Prototype:CheckFilters(item)
	for Filter, enabled in pairs(self.Filters) do
		if(enabled and (Filter(item, self) ~= true) ~= (enabled == -1)) then
			return nil
		end
	end
	return true
end

-- Set a filter for the object
function Prototype:SetFilter(filter, enabled, noUpdate)
	self.Filters[filter] = enabled
	if(not self.Init and not noUpdate) then cargBags:UpdateBags() end
end


-- The function for iterating the bags
-- it ensures that the buttons are placed in the right order
-- how they are in the Blizz bags
local iTable = {}
local iSort = function(a,b)
	return a.bagID == b.bagID and a.slotID < b.slotID or a.bagID < b.bagID
end

-- And now the function returning the iterator
local dummy = function() end
function Prototype:IterateButtons(func)
	local i = 1
	for button, _ in pairs(self.Buttons) do
		iTable[i] = button
		i = i + 1
	end
	while(iTable[i]) do
		iTable[i] = nil
		i = i + 1
	end
	sort(iTable, func or iSort)

	return next, iTable
end

-- Fetch new item data, into custom table or the item-one
function Prototype:RequestItemData(item, bagID, slotID)
	handler:LoadItemData(item or i, bagID or i.bagID, slotID or i.slotID)
	return item or i
end

-- Fetch the cooldown data
function Prototype:RequestCooldownData(item, bagID, slotID)
	handler:LoadCooldownData(item or i, bagID or i.bagID, slotID or i.slotID)
	return item or i
end

Prototype.SpawnPlugin = cargBags.SpawnPlugin
Prototype.GetHandler = cargBags.GetHandler

cargBags.Bags = bags
cargBags.Handler = handlers
cargBags.TempButtons = tempButtons
cargBags.Objects = objects
cargBags.Plugins = plugins
cargBags.Version = GetAddOnMetadata("cargBags", "Version")