--[[doc
h1. API: Handler

*Description:*
	Provides item and container information for cargBags

*Pipe functions:*
*	num = .GetContainerNumSlots(bagID, slotID): Gets the passed container's number of slots
*	free, bagType = .GetContainerNumFreeSlots(bagID, slotID): Gets the passed container's free slots and bag type
*	link = .GetInventoryItemLink(invID): Get the link of the selected bag
*	texture = .GetInventoryItemTexture(invID): Get the icon of the selected bag
*	num, full = .GetNumBankSlots(): returns the number of purchased bank slots and if you can buy more
*	copper = .GetMoney(): returns the player's money in copper

*Callback functions:*
*	:Enabled(): called when this handler was enabled
*	:Disable(): called when this handler was disabled
*	template = :GetButtonTemplate(bagID, slotID): Return the needed template for this button
*	button = :CreateButton(template, name): Return an item button with the passed name and template type
*	:LoadItemData(item, bagID, slotID, onlyLock): Write all available item data into the passed table
*	:LoadCooldownData(item, bagID, slotID): Write all available item cooldown data into the passed table
*	.BackSlotButton_OnEnter(button): called when a bag slot is entered by the mouse cursor
*	success = .PutItemInBag(bagID, slotID): Puts the item from the cursor in the selected container
*	PickupBagFromSlot(bagID, slotID): called when a bag should be picked from its bag slot
doc]]

local handler = CreateFrame"Frame"
handler.GetContainerNumSlots = GetContainerNumSlots
handler.GetContainerNumFreeSlots = GetContainerNumFreeSlots
handler.GetContainerItemInfo = GetContainerItemInfo
handler.GetContainerItemLink = GetContainerItemLink
handler.PutItemInBag = PutItemInBag
handler.BagSlotButton_OnEnter = BagSlotButton_OnEnter
handler.PickupBagFromSlot = PickupBagFromSlot
handler.GetInventoryItemLink = GetInventoryItemLink
handler.GetInventoryItemTexture = GetInventoryItemTexture
handler.GetNumBankSlots = GetNumBankSlots
handler.GetMoney = GetMoney

handler:SetScript("OnEvent", function(self, event, ...) cargBags:UpdateBags(event, ...) end)

-- Enable the handler ...
function handler:Enable()
	self:RegisterEvent"PLAYER_LOGIN"
	self:RegisterEvent"BAG_UPDATE"
	self:RegisterEvent"ITEM_LOCK_CHANGED"
	self:RegisterEvent"BAG_UPDATE_COOLDOWN"
	self:RegisterEvent"BANKFRAME_OPENED"
	self:RegisterEvent"PLAYERBANKSLOTS_CHANGED"
	self:RegisterEvent"PLAYERBANKBAGSLOTS_CHANGED"
	self:RegisterEvent"BAG_CLOSED"
	self:RegisterEvent"PLAYER_MONEY"
end

-- ... and disable it
handler.Disable = handler.UnregisterAllEvents

-- Which button template is needed?
function handler:GetButtonTemplateName(bagID, slotID)
	if(bagID == BANK_CONTAINER) then
		return "BankItemButtonGenericTemplate"
	else
		return "ContainerFrameItemButtonTemplate"
	end
end

-- Create an item button with the defined template and name
function handler:CreateButton(template, name)
	local button = CreateFrame("Button", name, nil, template)
	button.Icon = _G[name.."IconTexture"]
	button.Cooldown = _G[name.."Cooldown"]
	button.NormalTexture = _G[name.."NormalTexture"]
	button:SetHeight(37)
	button:SetWidth(37)
	return button
end

-- Load the item data ...
function handler:LoadItemData(i, onlyLock)
	i.texture, i.count, i.locked, i.quality, i.readable = GetContainerItemInfo(i.bagID, i.slotID)
	i.link = not onlyLock and GetContainerItemLink(i.bagID, i.slotID)

	if(not onlyLock and i.link) then
		i.name,i.link,i.rarity,i.level,i.minLevel,i.type,i.subType,i.stackCount,i.equipLoc = GetItemInfo(i.link)
	else
		i.name,i.link,i.rarity,i.level,i.minLevel,i.type,i.subType,i.stackCount,i.equipLoc = nil
	end
end

-- ... and cooldown data
function handler:LoadCooldownData(i)
	i.cdStart, i.cdFinish, i.cdEnable = GetContainerItemCooldown(i.bagID, i.slotID)
end

-- Register the handler
cargBags:RegisterHandler("Standard", handler)